;;;; -*- Mode: lisp; indent-tabs-mode: nil -*-
;;;
;;; advice.lisp - Implementation-independent defadvice facility
;;;
;;; Created 2005, Svein Ove Aas <svein.ove@aas.no>
;;;
;;; This code has been placed in the Public Domain.  All warranties are 
;;; disclaimed.

;;; Administrivia

(in-package :cl-user)

(eval-when (:compile-toplevel :load-toplevel :execute)
  (defpackage :swank-advice
    (:use :cl)
    (:nicknames :advice)
    (:export
     ;; Functions/macros
     #:defadvice
     #:add-advice
     #:remove-advice
     #:remove-all-advice
     #:delete-advice
     #:delete-all-advice
     #:update-symbol
     #:update-all
     #:call-next-advice
     ;; Conditions
     #:illegal-advice
     #:special-advice
     #:not-a-symbol)))

(in-package :swank-advice)

;;;;; Instructions:

;; Add advice with defadvice and/or add-advice; remove it with
;; remove/delete-advice.
;; If you know that a particular function has been changed, call update-symbol.
;; If you suspect that one may have been, but don't know which, call update-all
;; Use call-next-advice in :around functions, same as with :around methods -
;; but you must supply arguments.

;; For SLIME users: It updates automatically when you use the slime-mode
;; compilation/evaluation calls, but not if you call defun on the repl.
;; In this case, call advice:update-symbol or advice:update-all manually.

;; Caveats:
;;   - If you ever wrap a function by setting the fdefinition, updates will
;;     fail - possibly producing a loop.
;;   - Access to the exported functions is serialized, even for different
;;     symbols.
;;   - Advice name identities are determined by EQUAL, but you should probably
;;     use symbols.
;;   - Using advice will tend to clobber the argument list hints

;; TODO:
;; -  Hook into the backend implementation's compiler/evaluator, to make
;;    updates failsafe.
;; -  Copy the argument list from the base function, or (in the case of :around)
;;    the outermost around function.


(define-condition illegal-advice (error) ()
  (:report "Unspecified defadvice error"))

(define-condition special-advice (illegal-advice)
  ((operator :initarg :operator :reader operator))
  (:report (lambda (c s)
	     (format s "Cannot advice the special operator ~A" (operator c)))))

(define-condition macro-advice (illegal-advice)
  ((operator :initarg :operator :reader operator))
  (:report (lambda (c s)
	     (format s "Cannot advice the macro ~A" (operator c)))))

(define-condition undefined-advice (illegal-advice)
  ((operator :initarg :operator :reader operator))
  (:report (lambda (c s)
	     (format s "Cannot advice the undefined function ~A"
                     (operator c)))))

(define-condition not-a-symbol (illegal-advice)
  ((spec :initarg :spec :reader spec))
  (:report (lambda (c s)
	     (format s "Cannot advice the non-symbol ~S" (spec c)))))


(defmacro defadvice (symbol advice-name advice-type
		     lambda-list &body body)
  "Like add-advice, except that the symbol isn't evaluated; provided for
limited Lispworks compatibility."
  `(add-advice ',symbol ,advice-name ,advice-type
    (lambda ,lambda-list
      ,@body)))

(defun add-advice (symbol advice-name advice-type fn)
  "Defines new advice for a symbol, which must be bound to a function.
<defadvice symbol advice-name advice-type fn>
Advice for a single symbol is distinguished based on advice-name; a later
invocation of add-advice with the same (equal) advice-name will supersede an
earlier one.
There may be three types of advice for a given name: :before, :around and
:after. An :around advice may at its discretion call advice:call-next-advice.
The order in which multiple advice for a single function is invoked is
undefined."
  (cond ((listp symbol)
	 (defadvice-gf symbol advice-name advice-type fn))
	((not (symbolp symbol))
	 (error 'not-a-symbol :spec symbol))
	((special-operator-p symbol)
	 (error 'special-advice :operator symbol))
        ((macro-function symbol)
         (error 'macro-advice :operator symbol))
        ((not (fboundp symbol))
         (error 'undefined-advice :operator symbol))
	(t
	 (defadvice-standard symbol advice-name advice-type fn))))

(defun defadvice-gf (dspec name advice-type fn)
  (declare (ignore dspec name advice-type fn))
  (error "Not yet supported"))

(defclass advice ()
  ((symbol :initarg :symbol :reader advice-symbol)
   (before :initform (make-hash-table :test 'equal))
   (around :initform (make-hash-table :test 'equal))
   (after :initform (make-hash-table :test 'equal))
   (base-fn :initarg :base-fn :accessor base-fn)
   (result-fn :initform nil :accessor result-fn)))

(defun type->slot (type)
  (ecase type
    (:before 'before)
    (:around 'around)
    (:after 'after)))

(defvar *advice* (make-hash-table)) ; Indexed by symbol

(defun update-all ()
  "Rebuilds all adviced functions"
  (maphash (lambda (key value)
	     (declare (ignore value))
	     (%update-symbol key))
	   *advice*))


(defvar *next-around-advice* nil
  "Stack of around-style advice for call-next-advice")

(defun call-next-advice (&rest args)
  "Works pretty much like call-next-method, with the exception that you must
supply arguments."
  (when *next-around-advice*
    (apply (pop *next-around-advice*) args)))

(defun rebuild-advice (advice)
  (with-slots (symbol base-fn) advice
    (setf (fdefinition symbol) base-fn)
    (build-advice advice)))

(defun build-advice (advice)
  "Builds advice for a function, assuming there is none at the moment"
  (with-slots (before around after base-fn result-fn symbol)
      advice
    (setf base-fn (fdefinition symbol))
    (let (around-list)
      (push base-fn around-list)
      (maphash (lambda (key val)
                 (declare (ignore key))
                 (push val around-list))
               around)
      (setf result-fn
            (lambda (&rest args)
              (flet ((apply-val (key val)
                       (declare (ignore key))
                       (apply val args)))
                (maphash #'apply-val before)
                (prog1
                    (let ((*next-around-advice* around-list))
                      (apply 'call-next-advice args))
                  (maphash #'apply-val after)))))
      (setf (fdefinition symbol) result-fn))))

;;;;
;;;; Key assumption: If the result-fn recorded in the advice structure doesn't
;;;; match the fdefinition of the symbol, someone has recompiled it and we
;;;; should use the new fdefinition to build an entirely new function.
;;;;
;;;; This will fail - badly - if someone else tries to do the same thing we do.
;;;; Implementation-specific code might help.
;;;;

(defun %update-symbol (symbol)
  (let ((advice (gethash symbol *advice*)))
    (when advice
      (assert (eq (advice-symbol advice) symbol))
      (if (eq (fdefinition symbol)
              (result-fn advice))
          (rebuild-advice advice)
          (build-advice advice)))))

(defun advice-method (advice name advice-type)
  (gethash name (slot-value advice (type->slot advice-type))))

(defun (setf advice-method) (fn advice name advice-type)
  (setf (gethash name (slot-value advice (type->slot advice-type)))
	fn)
  (%update-symbol (advice-symbol advice)))

(defun defadvice-standard (dspec name advice-type fn)
  (let ((advice (gethash dspec *advice*)))
    (unless advice
      (setf advice (setf (gethash dspec *advice*)
			 (make-instance 'advice
					:base-fn (fdefinition dspec)
					:symbol dspec))))
    (setf (advice-method advice name advice-type) fn)))

(defmacro remove-advice (symbol advice-name advice-type)
  "Like delete-advice, but symbol is unevaluated"
  `(delete-advice ',symbol ,advice-name ,advice-type))

(defmacro remove-all-advice (symbol)
  "Like delete-all-advice, but symbol is unevaluated"
  `(delete-all-advice ',symbol))

(defun delete-advice (symbol advice-name advice-type)
  "Deletes a single advice from a single symbol"
  (let ((advice (gethash symbol *advice*)))
    (prog1
        (remhash advice-name (slot-value advice (type->slot advice-type)))
      (%update-symbol symbol))))

(defun delete-all-advice (symbol)
  "Deletes all advice from a single symbol"
  (let ((advice (gethash symbol *advice*)))
    (setf (fdefinition symbol)
          (base-fn advice))
    (remhash symbol *advice*)))

;;; Synchronization - our way. :)

(defvar *advice-lock* (swank-backend:make-lock :name "advice-lock"))

(defmacro serialize (function)
  `(defadvice ,function 'lock :around (&rest args)
    (swank-backend:call-with-lock-held *advice-lock*
     (lambda () (apply 'call-next-advice args)))))

;; To avoid recursive locks, %update-symbol doesn't lock.
(defun update-symbol (symbol)
  "Rebuilds advice for a single symbol"
  (%update-symbol symbol))

(serialize add-advice)
(serialize delete-advice)
(serialize delete-all-advice)
(serialize update-symbol)
(serialize update-all)
